"""
Logging to 2 files (log file + error log file), can be used both independently and in bifabrik pipelines.
The files use a CSV-like format so that they can be analysed using Spark
You can also log function calls including the arguments using the @logCalls decorator

    Examples
    --------
    
    a) Configure logging for use with bifabrik pipelines

    >>> import bifabrik as bif
    >>> 
    >>> # default = 'Files/BifabrikLog.log'
    >>> bif.config.log.logPath = '/log/log.csv'
    >>> 
    >>> # default = 'Files/BifabrikErrorLog.log'
    >>> bif.config.log.errorLogPath = '/log/error_log.csv'
    >>> 
    >>> # default = 'INFO'
    >>> bif.config.log.loggingLevel = 'DEBUG'
    >>> 
    >>> bif.fromCsv('Files/CsvFiles/annual-enterprise-survey-2021.csv') \
    >>>     .toTable('Survey2021').run()
    >>> 
    >>> bif.fromSql('''
    >>> SELECT CountryCode, FullName
    >>> FROM DimBranchZ LIMIT 3
    >>> ''').toTable('DimBranch2').run()

    b) Configure logging independently of the rest of the library and using the logger

    >>> from bifabrik.cfg.specific.LogConfiguration import LogConfiguration
    >>> from bifabrik.utils import log
    >>> from bifabrik.utils.log import logCalls
    >>>
    >>> cfg = LogConfiguration()
    >>> cfg.logPath = '/logFolder/log.csv'
    >>> cfg.errorLogPath = '/logFolder/error_log.csv'
    >>> cfg.loggingLevel = 'DEBUG'
    >>> logger = log.configureLogger(cfg)
    >>>
    >>> logger.info('test info log 1')
    >>> logger.info('test info log 2')
    >>> logger.error('error 1')
    >>> logger.error('error 2')
    >>> 
    >>> @logCalls
    >>> def loggedFunction(str):
    >>>   print(str)
    >>> 
    >>> loggedFunction('functionLog1')
    >>> loggedFunction('functionLog2')

"""

from bifabrik.cfg.specific.LogConfiguration import LogConfiguration
import bifabrik.utils.fsUtils as fsu

import logging
from functools import wraps
from inspect import getcallargs, getfullargspec
from collections.abc import Iterable
from collections import OrderedDict
from itertools import *
from bifabrik.utils.fsUtils import getDefaultLakehouseAbfsPath

__logger = None #logging.getLogger('bifabrik')
__errorLogHandler = None
__logHandler = None
__missingConfigWarningIssued = False

# called by bifabrik when starting a pipeline
def configureLogger(cfg: LogConfiguration) -> logging.Logger:
    """Initializes logging based on configuration"""
    
    global __logger
    global __errorLogHandler
    global __logHandler
    global __missingConfigWarningIssued

    if getDefaultLakehouseAbfsPath() is None:
        print('The logger cannot be setup because the notebook is not attached to a lakehouse. The logs would be written to the attached lakehouse under Files/.')
        __missingConfigWarningIssued = False
        return None
    
    # check if the logging cfg was modified
    # only do the setup if the log has not been set up yet or the cfg has been modified
    if (not cfg.modified) and (__logger is not None):
        if (len(__logger.handlers) > 0):
            return
    cfg.modified = False

    __logger = logging.getLogger('bifabrik')

    for handler in __logger.handlers[:]:
        __logger.removeHandler(handler)
    __logger.propagate = False

    if cfg.loggingEnabled == False:
        return

    level = logging.getLevelName(cfg.loggingLevel)
    __logger.setLevel(level)
    
    # logging to a different lakehouse cannot be supported now, because the default logger aims at a mounted directory in the environment
    # and other lakeouses are not in mounts
    # lhPath = fsu.getLakehousePath(lakehouse = cfg.logLakehouse, workspace = cfg.logWorkspace)
    # logPath = fsu.normalizeAbfsFilePath(cfg.logPath, lhPath)
    # errorLogPath = fsu.normalizeAbfsFilePath(cfg.errorLogPath, lhPath)

    logPath = fsu.normalizeFileApiPath(cfg.logPath)
    errorLogPath = fsu.normalizeFileApiPath(cfg.errorLogPath)

    __logHandler = logging.FileHandler(logPath, mode='a')
    __errorLogHandler = logging.FileHandler(errorLogPath, mode='a')
    
    __logHandler.setLevel(logging.DEBUG)
    __errorLogHandler.setLevel(logging.ERROR)

    log_format = logging.Formatter('%(asctime)s\t%(levelname)s\t%(message)s')
    __errorLogHandler.setFormatter(log_format)
    __logHandler.setFormatter(log_format)

    __logger.addHandler(__logHandler)
    __logger.addHandler(__errorLogHandler)
    return __logger


def getLogger(suppressWarnings = False):
    """Returns the 'bifabrik' logger if it has been initialized (using setLogger(LogConfiguration))
    Otherwise returns a new logger
    """
    global __logger
    global __errorLogHandler
    global __logHandler
    global __missingConfigWarningIssued

    if __missingConfigWarningIssued == False and __logger is None and suppressWarnings == False:
        print('Error: The bifabrik logger has not been properly configured. Please use configureLogger(cfg: LogConfiguration) first.')
        __missingConfigWarningIssued = True

    # this is done so that in case the log is not configured correcty, 
    # calls to the logger still pass and don't crash data loads
    if __logger is None:    
        __logger = logging.getLogger('bifabrik')

    return __logger
    #return logging.getLogger('bifabrik')

def logCalls(f):
    """A decorator to log every call to function (function name and arg values).
    The calls are logged at the INFO level.
    Based on https://gist.github.com/DarwinAwardWinner/1170921
    """
    @wraps(f)
    def func(*args, **kwargs):
        msg = __describe_call(f, *args, **kwargs)
        getLogger().info(msg)
        return f(*args, **kwargs)
    return func

def __flatten(l):
    """Flatten a list (or other iterable) recursively"""
    for el in l:
        if isinstance(el, Iterable) and not isinstance(el, str):
            for sub in __flatten(el):
                yield sub
        else:
            yield el

def __getargnames(func):
    """Return an iterator over all arg names, including nested arg names and varargs.
    Goes in the order of the functions argspec, with varargs and
    keyword args last if present."""
    #print(getfullargspec(func))
    (argnames, varargname, kwargname, defaults1, kwonlyargs1, kwonlydefaults1, annotations1) = getfullargspec(func)
    return chain(__flatten(argnames), filter(None, [varargname, kwargname]))

def __getcallargs_ordered(func, *args, **kwargs):
    """Return an OrderedDict of all arguments to a function.
    Items are ordered by the function's argspec."""
    argdict = getcallargs(func, *args, **kwargs)
    return OrderedDict((name, argdict[name]) for name in __getargnames(func))

def __describe_call(func, *args, **kwargs):
    res = "Calling %s("% func.__name__
    for argname, argvalue in __getcallargs_ordered(func, *args, **kwargs).items():
        res = res + ("\t%s = %s" % (argname, repr(argvalue)))
    res = res + ")"
    return res


from bifabrik.dst.DataDestination import DataDestination
from bifabrik.base.Pipeline import Pipeline
from bifabrik.cfg.TableDestinationConfiguration import TableDestinationConfiguration
from pyspark.sql.dataframe import DataFrame
from bifabrik.utils import fsUtils
import bifabrik.utils.log as lg
from bifabrik.utils import fsUtils
#from pyspark.sql.functions import col
from pyspark.sql.functions import *
from pyspark.sql.window import Window
import time
import datetime
import notebookutils.mssparkutils.fs
from bifabrik.utils import tableUtils as tu

class TableDestination(DataDestination, TableDestinationConfiguration):
    """Saves data to a lakehouse table.

    Examples
    --------
    > import bifabrik as bif
    >
    > bif.fromSql.query('SELECT * FROM SomeTable').toTable('DestinationTable1').run()
    """
    def __init__(self, pipeline: Pipeline, targetTableName: str):
        super().__init__(pipeline)
        TableDestinationConfiguration.__init__(self)
        self.__targetTableName = targetTableName
        self.__data = None
        self.__config = None
        self.__lhBasePath = None
        self.__identityColumn = None
        self.__insertDateColumn = None
        self.__tableExists = False
        self.__logger = None
        self.__tableLocation = None

    def __str__(self):
        return f'Table destination: {self.__targetTableName}'
    
    def execute(self, input: DataFrame) -> None:
        lgr = lg.getLogger()
        self.__logger = lgr
        self._error = None
        
        if input is None:
            msg = f'The table destination `{self.__targetTableName}` has no input; terminating'
            print(msg)
            lgr.warning(msg)
            return
        
        self.__data = input

        mergedConfig = self._pipeline.configuration.mergeToCopy(self)
        self.__config = mergedConfig
        self.__tableConfig = mergedConfig.destinationTable
        
        # print(f'dst tbl cfg')
        # print(f'increment {self.__tableConfig.increment}')
        # print(f'mergeKeyColumns {self.__tableConfig.mergeKeyColumns}')
        # print(f'snapshotKeyColumns {self.__tableConfig.snapshotKeyColumns}')
        # print(f'identityColumnPattern {self.__tableConfig.identityColumnPattern}')
        # print(f'insertDateColumn {self.__tableConfig.insertDateColumn}')
        # print(f'invalidCharactersInColumnNamesReplacement {self.__tableConfig.invalidCharactersInColumnNamesReplacement}')

        dstLh = mergedConfig.destinationStorage.destinationLakehouse
        dstWs = mergedConfig.destinationStorage.destinationWorkspace
        self.__lhBasePath = fsUtils.getLakehousePath(dstLh, dstWs)
        self.__lhMeta = fsUtils.getLakehouseMeta(dstLh, dstWs)
        self.__tableLocation = self.__lhBasePath + "/Tables/" + self.__targetTableName
        self.__tableExists = self.__tableExistsF()

        self.__replaceInvalidCharactersInColumnNames()
        self.__insertIdentityColumn()
        self.__insertInsertDateColumn()


        incrementMethod = mergedConfig.destinationTable.increment
        if incrementMethod is None:
            incrementMethod = 'overwrite'
        self.__incrementMethod = incrementMethod

        self.__resolveSchemaDifferences()
        
        self.__filterByWatermark()

        # if the target target table does not exist yet, just handle it as an overwrite
        if not(self.__tableExists):
            self.__overwriteTarget()
        elif incrementMethod == 'overwrite':
            self.__overwriteTarget()
        elif incrementMethod == 'append':
            self.__appendTarget()
        elif incrementMethod == 'merge':
            self.__mergeTarget()
        elif incrementMethod == 'snapshot':
            self.__snapshotTarget()
        else:
            raise Exception(f'Unrecognized increment type: {incrementMethod}')

        self._completed = True
        
    def __list_diff(self, first_list, second_list):
        diff = [item for item in first_list if item not in second_list]
        return diff
        
    def __replaceInvalidCharactersInColumnNames(self):
        replacement = self.__tableConfig.invalidCharactersInColumnNamesReplacement
        if replacement is None:
            return
        
        self.__data = (
            self.
            __data
            .select(
                [col(f'`{c}`').alias(
                    self.__sanitizeColumnName(c)) 
                    for c in self.__data.columns
                ]))

    def __sanitizeColumnName(self, colName):
        replacement = self.__tableConfig.invalidCharactersInColumnNamesReplacement
        if replacement is None:
            return colName
        
        invalids = " ,;{}()\n\t="
        name = colName        
        for i in invalids:
            name = name.replace(i, replacement)
        return name

    def __insertIdentityColumn(self):
        identityColumnPattern = self.__tableConfig.identityColumnPattern
        if identityColumnPattern is None:
            return
        
        lhName = self.__lhMeta.lakehouseName
        initID = 0
        identityColumn = self.__tableConfig.identityColumnPattern.format(tablename = self.__targetTableName, lakehousename = lhName)
        self.__identityColumn = identityColumn

        if self.__tableExists:
            query = f"SELECT MAX({identityColumn}) FROM {lhName}.{self.__targetTableName}"
            initID = self._spark.sql(query).collect()[0][0]
            # if the table exists, but is empty
            if initID is None:
                initID = 0
        
        # place the identity column at the beginning of the table
        schema = [obj[0]  for obj in self.__data.dtypes]
        schema.insert(0, identityColumn)
        df_ids  = self.__data.withColumn(identityColumn, row_number().over(Window.orderBy(schema[1])).cast('bigint')).select(schema)
        df_res = df_ids.withColumn(identityColumn,col(identityColumn) + initID)
        
        self.__data = df_res

    def __insertInsertDateColumn(self):
        insertDateColumn = self.__tableConfig.insertDateColumn
        if insertDateColumn is None:
            return
        
        self.__insertDateColumn = insertDateColumn
        ts = time.time()
        r = self.__data.withColumn(insertDateColumn, lit(ts).cast("timestamp"))
        self.__data = r

    def __overwriteTarget(self):
        self.__data.write.mode("overwrite").format("delta").option("overwriteSchema", "true").save(self.__tableLocation)

    def __appendTarget(self):
        self.__data.write.mode("append").format("delta").save(self.__tableLocation)

    def __resolveSchemaDifferences(self):
        lgr = lg.getLogger()

        if not(self.__tableExists):
            return
        if self.__incrementMethod == 'overwrite':
            return
        
        target_table = f'{self.__lhMeta.lakehouseName}.{self.__targetTableName}'
        df_old = self._spark.sql(f"SELECT * FROM {target_table} LIMIT 0")
        cols_new = self._spark.createDataFrame(self.__data.dtypes, ["new_name", "new_type"])
        cols_old = self._spark.createDataFrame(df_old.dtypes, ["old_name", "old_type"])
        compare = (
            cols_new.join(cols_old, cols_new.new_name == cols_old.old_name, how="outer")
            .fillna("")
        )
        difference = compare.filter(compare.new_type != compare.old_type)

        canAddColumns = self.__tableConfig.canAddNewColumns

        if canAddColumns:
            solvable = difference.where(difference.old_name == '')
            insolvable = difference.where(difference.old_name != '')
        else:
            solvable = difference.where(difference.old_type == '____none____')
            insolvable = difference

        if insolvable.count() > 0:
            err = f'Schema difference detected in table {target_table} that cannot be merged:'
            for r in insolvable.collect():
                ons = "N/A" if r.old_name == '' else r.old_name
                ots = "N/A" if r.old_type == '' else r.old_type
                nns = "N/A" if r.new_name == '' else r.new_name
                nts = "N/A" if r.new_type == '' else r.new_type
                err = err + f'\n> old_name: {ons}, old_type: {ots}'
                err = err + f', new_name: {nns}, new_type: {nts}'
            lgr.error(err)
            insolvable.show()
            raise Exception(err)

        for r in solvable.collect():
            tu.addTableColumnFromType(self.__lhMeta.lakehouseName, self.__targetTableName, r.new_name, r.new_type)


    def __mergeTarget(self):
        all_columns = self.__data.columns
        key_columns = list(map(lambda x: self.__sanitizeColumnName(x), self.__tableConfig.mergeKeyColumns))
        non_key_columns = self.__list_diff(self.__list_diff(all_columns, key_columns), [self.__identityColumn, self.__insertDateColumn])
        
        # print('key columns')
        # print(key_columns)
        # print('non-key columns')
        # print(non_key_columns)
        
        if len(key_columns) == 0:
            raise Exception('No key columns set for merge increment. Please set the mergeKeyColumns property in destinationTable configuration to the list of column names.')
        

        # todo: instead of SQL, use pyspark for cross-workspace ETL
        join_condition = " AND ".join([f"src.`{item}` = tgt.`{item}`" for item in key_columns])
        update_list = ", ".join([f"`{item}` = src.`{item}`" for item in non_key_columns])
        insert_list = ", ".join([f"`{item}`" for item in all_columns])
        insert_values = ", ".join([f"src.`{item}`" for item in all_columns])
        
        scd1_update = f"WHEN MATCHED THEN UPDATE SET \
            {update_list} "
        # table consisisting only of key columns - no need for the update clause
        if len(non_key_columns) == 0:
            scd1_update = ""

        src_view_name = f"src_{self.__lhMeta.lakehouseName}_{self.__targetTableName}"
        self.__data.createOrReplaceTempView(src_view_name)
        mergeDbRef = f'{self.__lhMeta.lakehouseName}.' 
        merge_sql = f"MERGE INTO {mergeDbRef}{self.__targetTableName} AS tgt \
            USING {src_view_name}  AS src \
            ON {join_condition} \
            {scd1_update}\
            WHEN NOT MATCHED THEN INSERT ( \
            {insert_list} \
            ) VALUES ( \
            {insert_values} \
            )\
            "
        self.__logger.info("----SCD1 MERGE SQL")
        self.__logger.info(merge_sql)
        self._spark.sql(merge_sql)

    def __snapshotTarget(self):
        # first delete the snapshot to be replaced
        key_columns = list(map(lambda x: self.__sanitizeColumnName(x), self.__tableConfig.snapshotKeyColumns))
        
        if len(key_columns) == 0:
            raise Exception('No key columns set for snapshot increment. Please set the snapshotKeyColumns property in destinationTable configuration to the list of column names.')
        
        join_condition = " AND ".join([f"src.`{item}` = tgt.`{item}`" for item in key_columns])

        join_column_list = ", ".join([f"`{item}`" for item in key_columns])
        
        src_view_name = f"src_{self.__lhMeta.lakehouseName}_{self.__targetTableName}"
        self.__data.createOrReplaceTempView(src_view_name)
        mergeDbRef = f'{self.__lhMeta.lakehouseName}.'

        merge_delete_sql = f"MERGE INTO {mergeDbRef}{self.__targetTableName} AS tgt \
            USING (SELECT DISTINCT {join_column_list} FROM {src_view_name})  AS src \
            ON {join_condition} \
            WHEN MATCHED THEN DELETE \
            "
        
        self.__logger.info("----SNAPSHOT DELETE SQL")
        self.__logger.info(merge_delete_sql)
        #print(merge_delete_sql)
        self._spark.sql(merge_delete_sql)
    
        # append the new rows
        self.__appendTarget()

    def __tableExistsF(self):
         """Checks if a table in the lakehouse exists.
         """
         fileExists = notebookutils.mssparkutils.fs.exists(self.__tableLocation)
         
        #  dstLh = self.__config.destinationStorage.destinationLakehouse
        #  dstWs = self.__config.destinationStorage.destinationWorkspace
        #  p = f'Tables/{self.__targetTableName}'
        #  print(f'Looking for table {p}')
        #  fileExists = fsUtils.fileExists(path = p, lakehouse = dstLh, workspace = dstWs)
         
        #  if fileExists:
        #      print('exists')
        #  else:
        #      print('does not exist')
         return fileExists
    
    def __filterByWatermark(self):
        if not(self.__tableExists):
            return
        elif self.__incrementMethod is None:
            return
        elif self.__incrementMethod == 'overwrite':
            return
        
        watermarkColumn = self.__tableConfig.watermarkColumn
        if watermarkColumn is None:
            return
        
        max_watermark_sql = f'SELECT MAX(`{watermarkColumn}`) FROM {self.__lhMeta.lakehouseName}.{self.__targetTableName}'
        # print(max_watermark_sql)
        max_watermark = self._spark.sql(max_watermark_sql).collect()[0][0]
        # print(f'max watermark: {max_watermark}')

        # if the table is empty or watermark is 0, take all the data
        if max_watermark is None:
            return
        if not max_watermark:
            return
        
        filter = f'{watermarkColumn} > "{max_watermark}"'
        self.__data = self.__data.filter(filter)
        # print(filter)
        # print(self.__data.count())


import json
from bifabrik.cfg.engine.ConfigContainer import ConfigContainer

from bifabrik.cfg.specific.LogConfiguration import LogConfiguration
from bifabrik.cfg.specific.CsvConfiguration import CsvConfiguration
from bifabrik.cfg.specific.JsonConfiguration import JsonConfiguration
from bifabrik.cfg.specific.ExcelConfiguration import ExcelConfiguration
from bifabrik.cfg.specific.FileSourceConfiguration import FileSourceConfiguration
from bifabrik.cfg.specific.SourceStorageConfiguration import SourceStorageConfiguration
from bifabrik.cfg.specific.DestinationStorageConfiguration import DestinationStorageConfiguration
from bifabrik.cfg.specific.MetadataStorageConfiguration import MetadataStorageConfiguration
from bifabrik.cfg.specific.TableConfiguration import TableConfiguration
from bifabrik.cfg.specific.SecurityConfiguration import SecurityConfiguration

import notebookutils.mssparkutils.fs

class CompleteConfiguration(ConfigContainer):
    """
    Fluent API for configuration
    
    You can set configuration directly through setter methods like .delimiter(';')[.further options...]
    or
    .option(name, value = None) gets / sets a named config property as in
    .option('delimiter', ';') #setter
    .option('delimiter') #getter
    """
    def __init__(self):
        self.__log = LogConfiguration()
        self.__csv = CsvConfiguration()
        self.__json = JsonConfiguration()
        self.__excel = ExcelConfiguration()
        self.__fileSource = FileSourceConfiguration()
        self.__sourceStorage = SourceStorageConfiguration()
        self.__destinationStorage = DestinationStorageConfiguration()
        self.__metadataStorage = MetadataStorageConfiguration()
        self.__destinationTable = TableConfiguration()
        self.__security = SecurityConfiguration()
        super().__init__()
        

    @property
    def csv(self) -> CsvConfiguration:
        """CSV files settings"""
        return self.__csv
    
    @property
    def json(self) -> JsonConfiguration:
        """JSON files settings"""
        return self.__json
    
    @property
    def excel(self) -> ExcelConfiguration:
        """Excel source files settings"""
        return self.__excel
    
    @property
    def log(self) -> LogConfiguration:
        """Logging settings"""
        return self.__log
    
    @property
    def fileSource(self) -> FileSourceConfiguration:
        """General file loading settings"""
        return self.__fileSource
    
    @property
    def sourceStorage(self) -> SourceStorageConfiguration:
        """The source lakehouse and warehouse reference used by pipelines"""
        return self.__sourceStorage

    @property
    def destinationStorage(self) -> DestinationStorageConfiguration:
        """The destination lakehouse and warehouse reference used by pipelines"""
        return self.__destinationStorage
    
    @property
    def metadataStorage(self) -> MetadataStorageConfiguration:
        """The metadata and backup storage lakehouse reference"""
        return self.__metadataStorage
    
    @property
    def destinationTable(self) -> TableConfiguration:
        "Destination table settings (increment method, identity column, etc.)"
        return self.__destinationTable
    
    @property
    def security(self) -> SecurityConfiguration:
        return self.__security
    
    def serialize(self) -> str:
        """Serializes the configuration to a string"""
        res = {}
        rootAttribs = dir(self)
        noUnderscoreRootAttrNames = list(filter(lambda x: not x.startswith("_"), rootAttribs))
        for rootAttrName in noUnderscoreRootAttrNames:
            rootAttr = getattr(self, rootAttrName)
            attrType = type(rootAttr)
            if attrType.__bases__[0].__name__ == 'Configuration':
                res[rootAttrName] = rootAttr._explicitProps
        j = json.dumps(res, indent=4)
        return j

    @staticmethod
    def deserialize(json_data):
        """Deserializes a CompleteConfiguration from a string"""
        json_dict = json.loads(json_data)
        cfg = CompleteConfiguration()
        for cfgAttrName in json_dict:
            for cfgPropName in json_dict[cfgAttrName]:
                cfg.option(cfgPropName, json_dict[cfgAttrName][cfgPropName])
        return cfg
    
    def saveToFile(self, path):
        """Saves the configuration to a JSON file"""
        serialized = self.serialize()
        notebookutils.mssparkutils.fs.put(path, serialized, True)

    def loadFromFile(self, path):
        """Loads the configuration values from a JSON file"""
        json_data = notebookutils.mssparkutils.fs.head(path, 1024 * 1024 * 1024)
        deser = CompleteConfiguration.deserialize(json_data)
        self.merge(deser)
        return self
        


from bifabrik.src.DataSource import DataSource
from bifabrik.cfg.ExcelSourceConfiguration import ExcelSourceConfiguration
from pyspark.sql.dataframe import DataFrame
import pandas as pd
from bifabrik.utils import fsUtils
from bifabrik.utils import log

class ExcelSource(DataSource, ExcelSourceConfiguration):
    """Excel data source
    uses the pandas loaders
    supports configuration options from https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.read_excel.html
    
    Examples
    --------
    > import bifabrik as bif
    >
    > bif.fromExcel('ExcelData/factOrderLine.xlsx').sheetName('Sheet1').toTable('FactOrderLine').run()
    """
    
    def __init__(self, parentPipeline):
        super().__init__(parentPipeline)
        ExcelSourceConfiguration.__init__(self)
        self._path = ""
        self.__processed_abfs_paths = []
        self.__mergedConfig = None

    def __str__(self):
        return f'Excel source: {self._path}'
    
    def path(self, path: str):
        """Set the path (or pattern) to the source file.
        It searches the Files/ folder in the current lakehouse
        """
        self._path = path
        return self
    
    def execute(self, input):
        lgr = log.getLogger()
        mergedConfig = self._pipeline.configuration.mergeToCopy(self)
        self.__mergedConfig = mergedConfig

        srcLh = mergedConfig.sourceStorage.sourceLakehouse
        srcWs = mergedConfig.sourceStorage.sourceWorkspace

        # pandas seems to have a problem with absolute ABFS paths, so use /lakehouse/default/ if refering to the default lakehouse
        source_files = fsUtils.filePatternSearch(self._path, srcLh, srcWs, useImplicitDefaultLakehousePath = True)
        implicitLhPath = fsUtils.getLakehousePath(lakehouse = srcLh, workspace = srcWs, useImplicitDefaultLakehousePath = True)
        abfsLhPath = fsUtils.getLakehousePath(lakehouse = srcLh, workspace = srcWs, useImplicitDefaultLakehousePath = False)
        self.__processed_abfs_paths = [x.replace(implicitLhPath, abfsLhPath) for x in source_files]

        if len(source_files) == 0:
            return None
        
        separator = ', '
        lgr.info(f'Loading Excel files: [{separator.join(source_files)}]')
        
        #source_files = fsUtils.filePatternSearch(self._path)
        fileDfs = []
        for src_file in source_files:
            #csv_path = f'/lakehouse/default/{src_file}'
            pd_df = pd.read_excel(
                io = src_file, 
                sheet_name = mergedConfig.excel.sheetName,
                header = mergedConfig.excel.header,
                names = mergedConfig.excel.names,
                usecols = mergedConfig.excel.usecols,
                skiprows = mergedConfig.excel.skiprows,
                nrows = mergedConfig.excel.nrows,
                thousands = mergedConfig.excel.thousands,
                decimal = mergedConfig.excel.decimal
                )
            df = self._spark.createDataFrame(pd_df)
            fileDfs.append(df)
        if len(fileDfs) == 0:
            return None
        df = fileDfs[0]
        if len(fileDfs) > 1:
            for i in range(1, len(fileDfs)):
                df = df.union(fileDfs[i])
        
        self._result = df
        self._completed = True

    def cleanup(self):
        if self.__mergedConfig.fileSource.moveFilesToArchive:
            fsUtils.archiveFiles(files = self.__processed_abfs_paths, 
                                 archiveFolder = self.__mergedConfig.fileSource.archiveFolder, 
                                 filePattern = self.__mergedConfig.fileSource.archiveFilePattern, 
                                 lakehouse = self.__mergedConfig.sourceStorage.sourceLakehouse, 
                                 workspace = self.__mergedConfig.sourceStorage.sourceWorkspace)


from bifabrik.cfg.engine.Configuration import Configuration
from bifabrik.cfg.engine.Configuration import CfgProperty

class JsonConfiguration(Configuration):
    """Configuration related to loading CSV files
    """
    def __init__(self):
        self._explicitProps = {}
        self.__primitivesAsString = None
        self.__prefersDecimal = None
        self.__allowComments = None
        self.__allowUnquotedFieldNames = None
        self.__allowSingleQuotes = None
        self.__allowNumericLeadingZero = None
        self.__allowBackslashEscapingAnyCharacter = None
        self.__jsonCorruptRecordsMode = None
        self.__jsonDateFormat = None
        self.__timestampFormat = None
        self.__multiLine = None
        self.__allowUnquotedControlChars = None
        self.__lineSep = None
        self.__samplingRatio = None

    @CfgProperty
    def primitivesAsString(self) -> str:
        """From spark:
        infers all primitive values as a string type. If None is set, it uses the default value, false.
        """
        return self.__primitivesAsString
    @primitivesAsString.setter(key='primitivesAsString')
    def primitivesAsString(self, val):
        self.__primitivesAsString = val

    @CfgProperty
    def prefersDecimal(self) -> str:
        """From spark:
        infers all floating-point values as a decimal type. If the values do not fit in decimal, then it infers them as doubles. If None is set, it uses the default value, false.
        """
        return self.__prefersDecimal
    @prefersDecimal.setter(key='prefersDecimal')
    def prefersDecimal(self, val):
        self.__prefersDecimal = val

    @CfgProperty
    def allowComments(self) -> str:
        """From spark:
        ignores Java/C++ style comment in JSON records. If None is set, it uses the default value, false.
        """
        return self.__allowComments
    @allowComments.setter(key='allowComments')
    def allowComments(self, val):
        self.__allowComments = val

    @CfgProperty
    def allowUnquotedFieldNames(self) -> str:
        """From spark:
        allows unquoted JSON field names. If None is set, it uses the default value, false.
        """
        return self.__allowUnquotedFieldNames
    @allowUnquotedFieldNames.setter(key='allowUnquotedFieldNames')
    def allowUnquotedFieldNames(self, val):
        self.__allowUnquotedFieldNames = val

    @CfgProperty
    def allowSingleQuotes(self) -> str:
        """From spark:
        allows single quotes in addition to double quotes. If None is set, it uses the default value, true.
        """
        return self.__allowSingleQuotes
    @allowSingleQuotes.setter(key='allowSingleQuotes')
    def allowSingleQuotes(self, val):
        self.__allowSingleQuotes = val

    @CfgProperty
    def allowNumericLeadingZero(self) -> str:
        """From spark:
        allows leading zeros in numbers (e.g. 00012). If None is set, it uses the default value, false.
        """
        return self.__allowNumericLeadingZero
    @allowNumericLeadingZero.setter(key='allowNumericLeadingZero')
    def allowNumericLeadingZero(self, val):
        self.__allowNumericLeadingZero = val

    @CfgProperty
    def allowBackslashEscapingAnyCharacter(self) -> str:
        """From spark:
        allows accepting quoting of all character using backslash quoting mechanism. If None is set, it uses the default value, false.
        """
        return self.__allowBackslashEscapingAnyCharacter
    @allowBackslashEscapingAnyCharacter.setter(key='allowBackslashEscapingAnyCharacter')
    def allowBackslashEscapingAnyCharacter(self, val):
        self.__allowBackslashEscapingAnyCharacter = val

    @CfgProperty
    def jsonCorruptRecordsMode(self) -> str:
        """From spark:
        allows a mode for dealing with corrupt records during parsing. If None is
set, it uses the default value, PERMISSIVE.

PERMISSIVE: when it meets a corrupted record, puts the malformed string into a field configured by columnNameOfCorruptRecord, and sets malformed fields to null. To keep corrupt records, an user can set a string type field named columnNameOfCorruptRecord in an user-defined schema. If a schema does not have the field, it drops corrupt records during parsing. When inferring a schema, it implicitly adds a columnNameOfCorruptRecord field in an output schema.

DROPMALFORMED: ignores the whole corrupted records.

FAILFAST: throws an exception when it meets corrupted records.
        """
        return self.__jsonCorruptRecordsMode
    @jsonCorruptRecordsMode.setter(key='jsonCorruptRecordsMode')
    def jsonCorruptRecordsMode(self, val):
        self.__jsonCorruptRecordsMode = val

    @CfgProperty
    def jsonDateFormat(self) -> str:
        """From spark:
        sets the string that indicates a date format. Custom date formats follow the formats at datetime pattern. # noqa This applies to date type. If None is set, it uses the default value, yyyy-MM-dd.
        """
        return self.__jsonDateFormat
    @jsonDateFormat.setter(key='jsonDateFormat')
    def jsonDateFormat(self, val):
        self.__jsonDateFormat = val

    @CfgProperty
    def timestampFormat(self) -> str:
        """From spark:
        sets the string that indicates a timestamp format. Custom date formats follow the formats at datetime pattern. # noqa This applies to timestamp type. If None is set, it uses the default value, yyyy-MM-dd'T'HH:mm:ss[.SSS][XXX].
        """
        return self.__timestampFormat
    @timestampFormat.setter(key='timestampFormat')
    def timestampFormat(self, val):
        self.__timestampFormat = val

    @CfgProperty
    def multiLine(self) -> str:
        """From spark:
        parse one record, which may span multiple lines, per file. If None is set, it uses the default value, false.
        """
        return self.__multiLine
    @multiLine.setter(key='multiLine')
    def multiLine(self, val):
        self.__multiLine = val

    @CfgProperty
    def allowUnquotedControlChars(self) -> str:
        """From spark:
        allows JSON Strings to contain unquoted control characters (ASCII characters with value less than 32, including tab and line feed characters) or not.
        """
        return self.__allowUnquotedControlChars
    @allowUnquotedControlChars.setter(key='allowUnquotedControlChars')
    def allowUnquotedControlChars(self, val):
        self.__allowUnquotedControlChars = val

    @CfgProperty
    def lineSep(self) -> str:
        """From spark:
        defines the line separator that should be used for parsing. If None is set, it covers all \r, \r\n and \n.
        """
        return self.__lineSep
    @lineSep.setter(key='lineSep')
    def lineSep(self, val):
        self.__lineSep = val

    @CfgProperty
    def samplingRatio(self) -> str:
        """From spark:
        defines fraction of input JSON objects used for schema inferring. If None is set, it uses the default value, 1.0.
        """
        return self.__samplingRatio
    @samplingRatio.setter(key='samplingRatio')
    def samplingRatio(self, val):
        self.__samplingRatio = val


from bifabrik.src.DataSource import DataSource
from pyspark.sql.dataframe import DataFrame as SparkDf

class SparkDfSource(DataSource):
    """Use spark dataframe as source
    
    Examples
    --------
    
    >>> import bifabrik as bif
    >>>
    >>> df = spark.read.format("csv").option("header","true").load("Files/CsvFiles/annual-enterprise-survey-2021.csv")
    >>> bif.fromSparkDf(df).toTable('Table1').run()
    """

    def __init__(self, parentPipeline, df: SparkDf):
        if not isinstance(df, SparkDf):
            raise Exception('The source is not a spark dataframe')
        
        super().__init__(parentPipeline)
        self.__sparkDf = df
    
    def __str__(self):
        return f'Spark DF source: {self._result}'
    
    def execute(self, input) -> SparkDf:
        self._result = self.__sparkDf
        self._completed = True
        return self._result


from bifabrik.tsf.DataTransformation import DataTransformation
from pyspark.sql.dataframe import DataFrame as SparkDf

class SparkDfTransformation(DataTransformation):
    """Use spark dataframe as source
    
    Examples
    --------
    
    >>> import bifabrik as bif
    >>>
    >>> (
    >>> bif
    >>>     .fromCsv("Files/CsvFiles/annual-enterprise-survey-2021.csv")
    >>>     .transformSparkDf(lambda df: df.withColumn('NewColumn', lit('NewValue')))
    >>>     .toTable("Survey2")
    >>>     .run()
    >>> )
    >>> 
    """

    def __init__(self, parentPipeline, func):
        super().__init__(parentPipeline)
        self.__func = func
    
    def __str__(self):
        return f'Spark DF transformation: {self.__func}'
    
    def execute(self, input) -> SparkDf:
        self._result = self.__func(input)
        self._completed = True
        return self._result


from bifabrik.cfg.engine.ConfigContainer import ConfigContainer
from bifabrik.cfg.DataDestinationConfigurationBase import DataDestinationConfigurationBase
from bifabrik.cfg.specific.TableConfiguration import TableConfiguration

class TableDestinationConfiguration(DataDestinationConfigurationBase):

    def __init__(self):
        self.__destinationTable = TableConfiguration()
        super().__init__()

    @property
    def destinationTable(self) -> TableConfiguration:
        "Destination table settings (increment method, identity column, etc.)"
        return self.__destinationTable


from bifabrik.dst.DataDestination import DataDestination
from bifabrik.base.Pipeline import Pipeline
from bifabrik.cfg.TableDestinationConfiguration import TableDestinationConfiguration
from pandas.core.frame import DataFrame as PandasDf
from pyspark.sql.dataframe import DataFrame as SparkDf
from bifabrik.utils import fsUtils
from bifabrik.utils import log

class PandasDfDestination(DataDestination):
    """Saves data to a pandas dataframe.

    Examples
    --------
    > import bifabrik as bif
    >
    > pandas_df = bif.fromSql('SELECT * FROM SomeTable').toPandasDf().run()
    """
    def __init__(self, pipeline: Pipeline):
        super().__init__(pipeline)

    def __str__(self):
        return f'Pandas DF destination'
    
    def execute(self, input: SparkDf) -> None:
        self._result = input.toPandas()
        self._error = None
        self._completed = True


# from pyspark.sql.session import SparkSession
# from pyspark.sql.dataframe import DataFrame
from bifabrik.base.Task import Task

class DataDestination(Task):
    def __init__(self, parentPipeline):
        super().__init__(parentPipeline)

    # def __init__(self, dataLoader, sourceDf: DataFrame):
    #     self._sourceDf = sourceDf
    #     self._loader = dataLoader
    #     dataLoader.destination = self
    #     self._spark = self._loader.spark

    def save(self) -> None:
        """Save the data to the destination. This is the "run" method at the end of the chain. 
        Returns the result of the last task, if any.
        """
        self._pipeline.execute()

    def run(self) -> any:
        """Save the data to the destination. This is the "run" method at the end of the chain. 
        Returns the result of the last task, if any.
        """
        return self._pipeline.execute()



import copy
from bifabrik.cfg.engine.Configuration import CfgProperty

class ConfigContainer:
    def __init__(self):
        # mapping cfg property names from inenr configurations
        # to the Configuration objects
        self.__propDict = dict()
        self.setCfgOptions()
    
    def setterFactory(self, attrName):
        def directSetter(val):
            #print(f'DirectSetter: {str(self)} - set {attrName} to {val}')
            self.option(attrName, val)
            return self
        return directSetter
    
    def setCfgOptions(self):
        rootAttribs = dir(self)
        noUnderscoreRootAttrNames = list(filter(lambda x: not x.startswith("_"), rootAttribs))
        
        # for adding config options from the partial configs
        if self.__doc__ is None:
            self.__doc__ = 'Configuration options:'
        else:
            self.__doc__ = self.__doc__ + '\n\nConfiguration options:'

        for rootAttrName in noUnderscoreRootAttrNames:
            rootAttr = getattr(self, rootAttrName)
            attrType = type(rootAttr)
            if attrType.__bases__[0].__name__ == 'Configuration':
                cfgPropNames = dir(attrType)
                for cfgpName in cfgPropNames:
                    cfgAttribute = getattr(attrType, cfgpName)
                    if isinstance(cfgAttribute, CfgProperty):
                        if cfgpName in self.__propDict:
                            matchingAttrTypes = list(filter(lambda x: str(type(x)) == str(type(rootAttr)), self.__propDict[cfgpName]))
                            if len(matchingAttrTypes) == 0:
                                self.__propDict[cfgpName].append(rootAttr)
                            #if str(type(self.__propDict[cfgpName])) != str(type(rootAttr)):
                            #    raise Exception(f'bifabrik configuration conflict: {cfgpName} is defined both in {self.__propDict[cfgpName]} ({type(self.__propDict[cfgpName])}) and {rootAttr} ({type(rootAttr)})')
                        else:
                            self.__propDict[cfgpName] = [rootAttr]

                        # create a direct setter property for overriding the value in a fluent API
                        # so that we don't need to use .option()
                        # (however, don't override existent properties while doint this - rather ommit some, they will still be available through .ption()_

                        hasA = hasattr(self, cfgpName)
                        if hasA == False:
                            
                            directSetter = self.setterFactory(cfgpName)
                            directSetter.__doc__ = f'{(cfgAttribute.__doc__ or "")} \n(from {rootAttrName} configuration)'
                            directSetter.__name__ = cfgpName
                            setattr(self, cfgpName, directSetter)

                        # (don't use this - the __doc__ is in the type)
                        # instanceAttr = getattr(attr, cfgp)
                        self.__doc__ = (self.__doc__ or "") + f'\n{cfgpName}: {cfgAttribute.__doc__}\n'

       
    def option(self, name, value = None):
        if not (name in self.__propDict):
            raise Exception(f'Configuration key not found: {name}.')
        if value is None:
            val = getattr(self.__propDict[name][0], name)
            return val
        #setattr(self.__propDict[name], name, value)
        for prop in self.__propDict[name]:
            setattr(prop, name, value)
        return self

    def mergeToCopy(self, other):
        """Meges another configuration into this one; the 'other' config container takes priority.
        Returns a new configuration without affecting the original"""
        cp = self.copy()
        return cp.merge(other)

    
    def merge(self, other):
        """Meges another configuration into this one; the 'other' config container takes priority"""
        for key in other.propDict:
            for configPart in other.propDict[key]:
                # that have an explicit value set in the source
                if key in configPart._explicitProps:
                    # and are among target attributes
                    if key in self.__propDict:
                        for localProp in self.__propDict[key]:
                            if(str(type(localProp)) == str(type(configPart))):
                                setattr(localProp, key, getattr(configPart, key))
                                #print(f'setting {key} of {str(type(localProp))} to {getattr(configPart, key)} based on {str(type(configPart))}')
        return self
    
    def copy(self):
        """Cretes a deep copy of the config container"""
        return copy.deepcopy(self)

    
    @property
    def propDict(self):
        """Dictionary mapping available configuration properties to the Configuration objects containing them.
        For internal use.
        """
        return self.__propDict


from bifabrik.cfg.engine.Configuration import Configuration
from bifabrik.cfg.engine.Configuration import CfgProperty

class SecurityConfiguration(Configuration):
    """Credentials configuration, currently used in the SharePointListSource.
    (Credentials aren't actually stored here, this configuration only refers to KeyVault secrets.)
    """
    def __init__(self):
        self._explicitProps = {}
        self.__keyVaultUrl = None
        self.__loginKVSecretName = None
        self.__passwordKVSecretName = None

    @CfgProperty
    def keyVaultUrl(self) -> str:
        """Key Vault URL - e.g. 'https://kv-fabrik.vault.azure.net/'
        """
        return self.__keyVaultUrl
    @keyVaultUrl.setter(key='keyVaultUrl')
    def keyVaultUrl(self, val):
        self.__keyVaultUrl = val

    @CfgProperty
    def loginKVSecretName(self) -> str:
        """Secret in the Key Vault that contains the username
        """
        return self.__loginKVSecretName
    @loginKVSecretName.setter(key='loginKVSecretName')
    def loginKVSecretName(self, val):
        self.__loginKVSecretName = val
    
    @CfgProperty
    def passwordKVSecretName(self) -> str:
        """Secret in the Key Vault that contains the password
        """
        return self.__passwordKVSecretName
    @passwordKVSecretName.setter(key='passwordKVSecretName')
    def passwordKVSecretName(self, val):
        self.__passwordKVSecretName = val

    


#from pyspark.sql.session import SparkSession
from pyspark.sql.dataframe import DataFrame
from bifabrik.dst.TableDestination import TableDestination
from bifabrik.dst.SparkDfDestination import SparkDfDestination
from bifabrik.dst.PandasDfDestination import PandasDfDestination
from bifabrik.base.Task import Task
#from typing import Any, Self

class DataTransformation(Task):
    """Generic data source. 
    Contains methods to initiate a DataDestination after you are done setting the transformation.
    bif.sourceInit.sourceOption1(A).srcOpt2(B).transformation1(X).transformation2(Y).toTable("Tab").destinationOption(X)....run()
    """
    def __init__(self, parentPipeline):
        super().__init__(parentPipeline)
        

    # def toDf(self) -> DataFrame:
    #     """Converts the data to a spark dataframe
    #     """
    #     pass

    def toTable(self, targetTableName: str) -> TableDestination:
        """Sets the destination table name (table in the current lakehouse)
        """
        dst = TableDestination(self._pipeline, targetTableName)
        return dst
    
    def toSparkDf(self) -> SparkDfDestination:
        """Sets the destination as a Spark DF (for further processing)
        """
        dst = SparkDfDestination(self._pipeline)
        return dst

    def toPandasDf(self) -> PandasDfDestination:
        """Sets the destination as a Pandas DF (for further processing)
        """
        dst = PandasDfDestination(self._pipeline)
        return dst
    
    def transformSparkDf(self, func):
        """Applies the givet function to transform the data as a Spark DF
        """
        from .SparkDfTransformation import SparkDfTransformation
        tsf = SparkDfTransformation(self._pipeline, func)
        return tsf
    
    def transformPandasDf(self, func):
        """Applies the givet function to transform the data as a Pandas DF
        """
        from .PandasDfTransformation import PandasDfTransformation
        tsf = PandasDfTransformation(self._pipeline, func)
        return tsf



"""
Delta table utilities - add / rename / remove columns
"""

import pyspark.sql.session as pss
from pyspark.sql.functions import lit, col

spark = pss.SparkSession.builder.getOrCreate()

def addTableColumnFromValue(databaseName: str, tableName: str, columnName: str, value):
    """Adds a new column to the delta table with the provided value.
    The `database` parameter refers to a lakehouse in the current workspace
    """
    # find out the data type
    df = spark.createDataFrame([('0','0')], ['____', '_____'])
    df = df.withColumn(columnName, lit(value))
    dts = df.dtypes
    dfColT = dts[2]
    type_name = dfColT[1]
    alterQuery = f'ALTER TABLE `{databaseName}`.`{tableName}` ADD columns (`{columnName}` {type_name})'
    updateQuery = f'UPDATE `{databaseName}`.`{tableName}` SET `{columnName}` = CAST(\'{value}\' AS {type_name})'

    print(alterQuery)
    spark.sql(alterQuery)
    print(updateQuery)
    spark.sql(updateQuery)

def addTableColumnFromType(databaseName: str, tableName: str, columnName: str, typeName: str):
    """Adds a new empty column to the delta table.
    The `database` parameter refers to a lakehouse in the current workspace
    """
    alterQuery = f'ALTER TABLE `{databaseName}`.`{tableName}` ADD columns (`{columnName}` {typeName})'
    print(alterQuery)
    spark.sql(alterQuery)

def renameTableColumn(databaseName: str, tableName: str, columnName: str, newColumnName: str):
    """Renames a column in a delta table"""
    alterPropsQuery = f'''ALTER TABLE `{databaseName}`.`{tableName}` SET TBLPROPERTIES (
    'delta.minReaderVersion' = '2',
    'delta.minWriterVersion' = '5',
    'delta.columnMapping.mode' = 'name'
    )
    '''
    print(alterPropsQuery)
    spark.sql(alterPropsQuery)

    alterQuery = f'ALTER TABLE `{databaseName}`.`{tableName}` RENAME COLUMN `{columnName}` TO `{newColumnName}`'
    print(alterQuery)
    spark.sql(alterQuery)

def dropTableColumn(databaseName: str, tableName: str, columnName: str):
    """Drops a column from a delta table"""
    alterPropsQuery = f'''ALTER TABLE `{databaseName}`.`{tableName}` SET TBLPROPERTIES (
    'delta.minReaderVersion' = '2',
    'delta.minWriterVersion' = '5',
    'delta.columnMapping.mode' = 'name'
    )
    '''
    print(alterPropsQuery)
    spark.sql(alterPropsQuery)

    alterQuery = f'ALTER TABLE `{databaseName}`.`{tableName}` DROP COLUMN `{columnName}`'
    print(alterQuery)
    spark.sql(alterQuery)


"""
PowerBI dataset scripting utilities - tabular model backup & restore and some dataset modifications.
"""

import bifabrik.utils.fsUtils as fsu
import bifabrik.utils.log as lg
import bifabrik as bif
import datetime
import sempy.fabric as spf
import pyspark.sql.session as pss
import notebookutils.mssparkutils.fs
from pyspark.sql.functions import col
import uuid

def backupDataset(sourceDataset: str, sourceWorkspace: str, targetFilePath = None):
    """Save the dataset definition to a TMSL script. This serves as a backup of the definition. If you damage your semantic model, you can restore it from TMSL using restoreDataset().

    Parameters:
        sourceDataset (str):    Name or ID of the dataset to backup
        sourceWorkspace (str):  Name or ID of the workspace where the dataset is located
        targetFilePath (str):   Path to which the backup will be saved. If this is not specified, the bifabrik configuration will be used to find the metadata storage
                                and construct the file pattern
                                (see
                                    bifabrik.config.metadataStorage.metadataWorkspaceName, 
                                    bifabrik.config.metadataStorage.metadataLakehouseName, 
                                    bifabrik.config.metadataStorage.tmslBackupPathPattern)
    
    Internally, this function uses the semantic link library: https://learn.microsoft.com/en-us/python/api/semantic-link-sempy/sempy.fabric?view=semantic-link-python#sempy-fabric-get-tmsl

    Examples
    --------
    
    This example saves a the dataset definition to Files/SM_Test_backup.json in the default lakehouse. The notebook that runs this needs to have a default lakehouse selected.
    
    >>> import bifabrik.utils.tmslUtils as tmsl
    >>> tmsl.backupDataset(sourceWorkspace = 'DataWS', sourceDataset = 'SM_Test', targetFilePath = 'Files/SM_Test_backup.json')

    This example uses the metadataStorage configuration of bifabrik to set the location for dataset backups.
    
    >>> import bifabrik as bif
    >>> import bifabrik.utils.tmslUtils as tmsl
    >>>
    >>> bif.config.metadataStorage.metadataWorkspaceName = 'Experimental'
    >>> bif.config.metadataStorage.metadataLakehouseName = 'Lakehouse1'
    >>>
    >>> tmsl.backupDataset(sourceWorkspace = 'DataWS', sourceDataset = 'SM_Test')

    The resulting backup path can look like this: abfss://6dcac488-.../Files/tmsl_backup/DataWS/SM_Test/SM_Test_2024_03_18_17_34_52_472325.json
    """
    cfg = bif.config.metadataStorage
    
    #print(cfg)
    #print(f'metadataWorkspaceName: {cfg.metadataWorkspaceName}')
    #print(f'metadataLakehouseName: {cfg.metadataLakehouseName}')
    #print(f'tmslBackupPathPattern: {cfg.tmslBackupPathPattern}')

    ct = datetime.datetime.now()
    cts = ct.strftime("%Y_%m_%d_%H_%M_%S_%f")
    
    if targetFilePath is None:
        metaLakehouseBasePath = fsu.getLakehousePath(lakehouse = cfg.metadataLakehouseName, workspace = cfg.metadataWorkspaceName)
        targetPathReplace = cfg.tmslBackupPathPattern.format(workspacename = sourceWorkspace, datasetname = sourceDataset, timestamp = cts)
        targetFilePath = fsu.normalizeAbfsFilePath(targetPathReplace, metaLakehouseBasePath)

    tmsl = spf.get_tmsl(dataset = sourceDataset, workspace = sourceWorkspace)
    
    print(f'Saving backup to {targetFilePath}')
    notebookutils.mssparkutils.fs.put(targetFilePath, tmsl)

def restoreDataset(sourceFilePath: str, targetDatasetName: str, targetWorkspaceName: str):
    """Restore a dataset from backup. For creating the backup, use the backupDataset() function.
    If a dataset with the target name already exists in the workspace, it will be overwritten.
    If it does not exist, a new dataset will be created.

    Parameters:
        sourceFilePath (str):       Relative Spark path to the backup created by backupDataset()
        targetDatasetName (str):    Name of the restored dataset
        targetWorkspaceName (str):  Name of the workspace where to place the dataset.
    
    Examples
    --------
    >>> import bifabrik.utils.tmslUtils as tmsl
    >>> tmsl.restoreDataset(sourceFilePath = 'Files/tmsl_backup/Experimental/DS_R1/DS_R1_2024_03_22_22_05_57_580830.json', targetDatasetName = 'DS_R1_restored', targetWorkspaceName = 'Experimental')
    """
    datasets_pandas = spf.list_datasets(workspace = targetWorkspaceName)
    spark = pss.SparkSession.builder.getOrCreate()
    datasets = spark.createDataFrame(datasets_pandas)

    target_id = str(uuid.uuid4())
    
    # look for the destination dataset. If it already exists, take its ID (replace target)
    # otherwise, we're creating a new dataset

    ds_filter = datasets.filter(col("Dataset Name") == targetDatasetName)
    create_new = True
    if ds_filter.count() > 0:
        target_id = ds_filter.collect()[0]['Dataset ID']
        create_new = False
    
    rdd = spark.read.text(sourceFilePath)
    tmsl_def = ""
    name_replaced = False
    id_replaced = False
    
    for l in rdd.collect():
        s = l[0]
        if s.strip().startswith('"name":') and name_replaced == False:
            # print('orig ' + s)
            s = f'"name": "{targetDatasetName}",'
            name_replaced = True
            # print(s)
        if s.strip().startswith('"id":') and id_replaced == False:
            # print('orig ' + s)
            s = f'"id": "{target_id}",'
            id_replaced = True
            # print(s)
        tmsl_def = tmsl_def + '\n' + s

    wrap_start = f"""{{
    "createOrReplace": {{
        "object": {{
        "database": "{targetDatasetName}"
        }},
        "database":"""
    wrap_end = " } }"
    tmsl_cmd = wrap_start + tmsl_def + wrap_end

    if create_new:
        print(f'Creating dataset {targetDatasetName} in workspace {targetWorkspaceName}')
    else:
        print(f'Overwriting dataset {targetDatasetName} in workspace {targetWorkspaceName}')
    
    spf.execute_tmsl(script = tmsl_cmd, workspace = targetWorkspaceName)


from bifabrik.src.DataSource import DataSource
from bifabrik.cfg.SharePointListSourceConfiguration import SharePointListSourceConfiguration
from pyspark.sql.dataframe import DataFrame
from bifabrik.utils import log
from shareplum import Site, folder
from shareplum import Office365
from shareplum.site import Version
import notebookutils.mssparkutils.credentials

class SharePointListSource(DataSource, SharePointListSourceConfiguration):
    """Loads data from an Office365 SharePoint list. You will need to configure the credentials for SharePoint.
    
    Examples
    --------
    >>> import bifabrik as bif
    >>> 
    >>> bif.config.security.keyVaultUrl = 'https://kv-fabrik.vault.azure.net/'
    >>> bif.config.security.loginKVSecretName = 'SharePointLogin'
    >>> bif.config.security.passwordKVSecretName = 'SharePointPwd'
    >>> 
    >>> bif.fromSharePointList('https://fabrik.sharepoint.com/sites/BusinessIntelligence', 'CustomerList') \
    >>>     .toTable('Customers').run()
    """
    
    def __init__(self, parentPipeline, siteUrl: str, listName: str):
        super().__init__(parentPipeline)
        SharePointListSourceConfiguration.__init__(self)
        self.__siteUrl = siteUrl
        self.__listName = listName
        self.__mergedConfig = None

    def __str__(self):
        return f'SharePointListSource source: {self.__siteUrl}/Lists/{self.__listName}'
    
    def execute(self, input):
        lgr = log.getLogger()
        mergedConfig = self._pipeline.configuration.mergeToCopy(self)
        self.__mergedConfig = mergedConfig

        securityConfig = mergedConfig.security

# try:
#                 if tsk.completed == True:
#                     if tsk.error != None:
#                         raise Exception(tsk.error)
#                     prevResult = tsk.result
#                     continue
                
#                 if tsk.error != None:
#                     raise Exception(tsk.error)
                
#                 lgr.info(f'Executing {tsk}')
#                 tsk.execute(prevResult)
#                 prevResult = tsk.result
#             except Exception as e:
#                 lgr.exception(e)
#                 raise e

        kvLogin = None
        kvPwd = None
        try:
            kvLogin = notebookutils.mssparkutils.credentials.getSecret(securityConfig.keyVaultUrl, securityConfig.loginKVSecretName)
        except Exception as e:
            lgr.exception(e)
            raise Exception(f'Failed to fetch login secret \'{securityConfig.loginKVSecretName}\' from Key Vault \'{securityConfig.keyVaultUrl}\', see error log for details')
        try:
            kvPwd = notebookutils.mssparkutils.credentials.getSecret(securityConfig.keyVaultUrl, securityConfig.passwordKVSecretName)
        except Exception as e:
            lgr.exception(e)
            raise Exception(f'Failed to fetch password secret \'{securityConfig.passwordKVSecretName}\' from Key Vault \'{securityConfig.keyVaultUrl}\', see error log for details')
        
        ixSites = self.__siteUrl.index('/sites/')
        sharePointServerUrl = self.__siteUrl[:ixSites]

        authcookie = Office365(sharePointServerUrl, username=kvLogin, password=kvPwd).GetCookies()
        site = Site(self.__siteUrl, version=Version.v365, authcookie=authcookie)
        sp_list = site.List(self.__listName)
        sp_data = sp_list.GetListItems()
        sp_df = self._spark.createDataFrame(sp_data)
        
        self._result = sp_df
        self._completed = True